module mpi_helmholz_1d_cg_mod
!
! 1D Helmholz equation CG solver 
! for ASLFT2011 (2011/03/12-?)
! Ken-Ichi Ishikawa
! $Id: mpi_helmholz_1d_cg.F90,v 1.2 2011/03/09 09:56:41 ishikawa Exp $
!
! NO WARRANTY
!
  use mpi
  implicit none
  integer, parameter :: DP = KIND(1.0d0)
  integer, parameter :: NSITE = 100
  integer, save :: NLSITE                ! number of local sites
  integer, save :: N_MPI                 ! number of MPI processes
  integer, save :: RANK                  ! MPI rank number
  integer, save :: UP_RANK               !   Upward site MPI rank number
  integer, save :: DN_RANK               ! Downward site MPI rank number

contains

subroutine init_mpi_env()
!
! Initialize MPI environment, local lattice size, and node RANK relations
!
  implicit none
  integer :: ierr

  call MPI_INIT(ierr)
  if (MPI_SUCCESS /= ierr) then
    write(*,'("MPI_INIT ERROR STOP!.")')
    stop
  endif
  call MPI_COMM_RANK(MPI_COMM_WORLD,RANK,ierr)
  if (MPI_SUCCESS /= ierr) then
    write(*,'("MPI_COMM_RANK ERROR STOP!.")')
    stop
  endif
  call MPI_COMM_SIZE(MPI_COMM_WORLD,N_MPI,ierr)
  if (MPI_SUCCESS /= ierr) then
    write(*,'("MPI_COMM_SIZE ERROR STOP!.")')
    stop
  endif

  if (1 == N_MPI) then
    write(*,'("Single MPI process is not supported. N_MPI > 1, N_MPI=",I3)')N_MPI
    call stop_mpi_env()
    stop
  endif
  if (0 /= mod(N_MPI,2)) then
    write(*,'("Number of MPI Processes should an even number. N_MPI=",I3)')N_MPI
    call stop_mpi_env()
    stop
  endif

  if (0 /= mod(NSITE,N_MPI)) then
    write(*,'("Number of MPI Processes should exactly divides NSITE. NSITE=",I10," N_MPI=",I3)')NSITE,N_MPI
    call stop_mpi_env()
    stop
  endif

  !
  ! Set up local lattice size/rank structure
  !
  NLSITE  = NSITE/N_MPI
  UP_RANK = mod(RANK+1      ,N_MPI)
  DN_RANK = mod(RANK-1+N_MPI,N_MPI)

  return
end subroutine

subroutine stop_mpi_env()
!
! Finalize MPI enviroment
!
  implicit none
  integer :: ierr
  call MPI_FINALIZE(ierr)
  if (MPI_SUCCESS /= ierr) then
    write(*,'("MPI_FINALIZE ERROR STOP!.")')
    stop
  endif
  return
end subroutine

subroutine assign_mult_helmhz_op(k2,v,p)
!
! Return v = -Laplace(p) + k2 p
!
  implicit none
  real(DP), intent(in) :: k2
  real(DP), intent(out)   :: v(0:NLSITE+1)
  real(DP), intent(inout) :: p(0:NLSITE+1)
  integer :: i,ierr
  integer :: istat(MPI_STATUS_SIZE)

  !
  ! set periodic boundary condition on the ghost sites.
  !
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  if (mod(RANK,2) == 0) then

    call MPI_SEND(p(NLSITE)  ,1,MPI_REAL8,UP_RANK,1,MPI_COMM_WORLD,ierr)
    call MPI_SEND(p(1)       ,1,MPI_REAL8,DN_RANK,2,MPI_COMM_WORLD,ierr)

    call MPI_RECV(p(0)       ,1,MPI_REAL8,DN_RANK,3,MPI_COMM_WORLD,istat,ierr)
    call MPI_RECV(p(NLSITE+1),1,MPI_REAL8,UP_RANK,4,MPI_COMM_WORLD,istat,ierr)

  else

    call MPI_RECV(p(0)       ,1,MPI_REAL8,DN_RANK,1,MPI_COMM_WORLD,istat,ierr)
    call MPI_RECV(p(NLSITE+1),1,MPI_REAL8,UP_RANK,2,MPI_COMM_WORLD,istat,ierr)
    
    call MPI_SEND(p(NLSITE)  ,1,MPI_REAL8,UP_RANK,3,MPI_COMM_WORLD,ierr)
    call MPI_SEND(p(1)       ,1,MPI_REAL8,DN_RANK,4,MPI_COMM_WORLD,ierr)

  endif 
  
  ! 
  ! Apply Laplacian on p
  ! 
  do i=1,NLSITE
    v(i) = -p(i+1)+(2.0_DP+k2)*p(i)-p(i-1)
  enddo

  return
end subroutine

function abs2(v) result(nrm2)
!
! return squared vector norm
!
  implicit none
  real(DP), intent(in) :: v(0:NLSITE+1)
  real(DP) :: nrm2,rtmp
  integer :: i,ierr
  nrm2 = 0.0_DP
  rtmp = 0.0_DP
  do i=1,NLSITE
    rtmp = rtmp + v(i)**2
  enddo
  call MPI_ALLREDUCE(rtmp,nrm2,1,MPI_REAL8,MPI_SUM,MPI_COMM_WORLD,ierr)
  return
end function

function prod(p,q) result(prd)
!
! return inner product of <p|q>
!
  implicit none
  real(DP), intent(in) :: p(0:NLSITE+1)
  real(DP), intent(in) :: q(0:NLSITE+1)
  real(DP) :: prd,rtmp
  integer :: i,ierr
  prd = 0.0_DP
  rtmp = 0.0_DP
  do i=1,NLSITE
    rtmp = rtmp + p(i)*q(i)
  enddo
  call MPI_ALLREDUCE(rtmp,prd,1,MPI_REAL8,MPI_SUM,MPI_COMM_WORLD,ierr)
  return
end function

end module

program mpi_helmholz_1d_cg
  use mpi_helmholz_1d_cg_mod
  implicit none
  character(256) :: str
  real(DP), allocatable :: b(:)  ! = rho(:)   charge
  real(DP), allocatable :: x(:)  ! = phi(:)    field
  real(DP), allocatable :: r(:)  ! CG residual
  real(DP), allocatable :: p(:)  ! CG working vec
  real(DP), allocatable :: q(:)  ! CG working vec
  real(DP) :: k2
  real(DP) :: err,tol
  real(DP) :: rho0,rho1,alpha,beta  ! CG algorithm scalars
  integer :: i,iter,NITER,igx,ipx,ierr

  k2 = 0.1_DP
  NITER = 1000
  tol = EPSILON(1.0_DP)*10

  call init_mpi_env()

  if (0 == RANK) then
    write(*,'("# Solve 1D Helmholz eq")')
    write(*,'("#  KAPPA^2 = ",F9.6)')k2
    write(*,'("#   NSITE  = ",I4)')NSITE
    write(*,'("#   N_MPI  = ",I4)')N_MPI
    write(*,'("#  NLSITE  = ",I4)')NLSITE
    write(*,'("#   CG TOL = ",E22.15)')tol
    write(*,'("#ITER         |R|")')
  endif

  allocate(b(0:NLSITE+1))
  allocate(x(0:NLSITE+1))
  allocate(r(0:NLSITE+1))
  allocate(p(0:NLSITE+1))
  allocate(q(0:NLSITE+1))

  r(:) = 0.0_DP
  p(:) = 0.0_DP
  q(:) = 0.0_DP
  r(:) = 0.0_DP

  !
  ! set charge at site# = 1
  !
  b(:) = 0.0_DP
  igx = 1                     ! global index
  i = mod(igx-1,NLSITE)+1     ! local index
  ipx = (igx-1)/NLSITE        ! MPI RANK
  if (ipx == RANK) then
    b(igx) = 1.0_DP
  endif

  !
  ! set initial solution to zero
  !
  x(:) = 0.0_DP

  !
  ! CG algorithm
  !

  iter = 0

  call assign_mult_helmhz_op(k2,q,x)         ! q = A x
  r(1:NLSITE) = b(1:NLSITE) - q(1:NLSITE)    ! r = b - q = b - A x (initial resdual)
  p(1:NLSITE) = r(1:NLSITE)                  ! p = r
  rho0 = abs2(r)                             ! rho0 = <r|r> = |r|^2

  err = sqrt(rho0)
  if (0 == RANK)  &
 &  write(*,'(I3,2ES24.15)')iter,err

  do 
    iter = iter + 1

    call assign_mult_helmhz_op(k2,q,p)  ! q = A p

    alpha = rho0/prod(p,q)

    x(1:NLSITE) = x(1:NLSITE) + alpha * p(1:NLSITE)
    r(1:NLSITE) = r(1:NLSITE) - alpha * q(1:NLSITE)

    rho1 = abs2(r)                    ! rho1 = <r|r> = |r|^2
    err = sqrt(rho1)
    if (0 == RANK)  &
 &    write(*,'(I3,2ES24.15)')iter,err
    if (err < tol) exit
    if (iter > NITER) then
      if (0 == RANK)  &
 &      write(*,'("CG algorithm did not converge.",I4)')iter
      exit
    endif

    beta = rho1/rho0
    rho0 = rho1

    p(1:NLSITE) = r(1:NLSITE) + beta * p(1:NLSITE)

  enddo

  !
  ! check true residual
  !
  call assign_mult_helmhz_op(k2,q,x)        ! q = A x
  r(1:NLSITE) = b(1:NLSITE) - q(1:NLSITE)   ! r = b - q = b - A x
  err = sqrt(abs2(r))
  if (0 == RANK) then
    write(*,'("#ITER      TRUE RESIDUAL               TOL")')
    write(*,'("#",I3,2ES24.15)')iter,err,tol
  endif


  !
  ! print out solution and residual components
  !
  if (0 == RANK) then
    write(*,*)
    write(*,*)
    write(*,'("# i            X(i)                    R(i)          local-index   RANK")')
  endif

  !
  ! gather results to master node
  !
  do igx = 1,NSITE
    i = mod(igx-1,NLSITE)+1
    ipx = (igx-1)/NLSITE

    if (ipx == RANK) then
      write(str,'(I3,2ES24.15,I8,I11)')igx,x(i),r(i),i,RANK
    endif

    call MPI_BCAST(str,LEN(str),MPI_CHARACTER,ipx,MPI_COMM_WORLD,ierr)

    if (0 == RANK) then
      write(*,*)TRIM(str)
    endif
  enddo

  deallocate(b,x,r,p,q)

  call stop_mpi_env()

  stop
end program
